/*
 * Common display-related functions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "display.h"



/* Return the line data position for the given pixel. */

int get_position(display_config_t *cfg, int x)
{
    int cell = x / cfg->cell_size, offset = x % cfg->cell_size;
    int pos = (cell / 2) * cfg->cell_size + offset;

    return cell % 2 ? pos + cfg->line_length / 2 : pos;
}

/* Provide a pattern to test the line data. */

void test_linedata(display_config_t *cfg)
{
    int x, y;
    uint8_t *linedata = cfg->screen_start;

    for (y = 0; y < cfg->line_count; y++)
    {
        for (x = 0; x < cfg->line_length; x++)
        {
            /* Pixel: I0RRGGBB = Y0YYYYXX */

            linedata[get_position(cfg, x)] = (x % 2) ?
                    (((y / (cfg->line_count / 32)) & 0b1) << 7) |
                    (((y / (cfg->line_count / 16)) & 0b1111) << 2) |
                    ((x / (cfg->line_length / 4)) & 0b11) :
                0x00;
        }

        linedata += cfg->line_length;

        if (linedata >= cfg->screen_limit)
            linedata -= cfg->screen_size;
    }
}

/* Copying from/to the display to/from a backing store. */

void copy_display(display_config_t *cfg, uint8_t *store, int width, int height,
                  int x, int y, int key, int to_display)
{
    copy_display_section(cfg, store, width, height, 0, 0, width, height, x, y,
                         key, to_display);
}

/* Copying from/to the display to/from a backing store region. */

void copy_display_section(display_config_t *cfg, uint8_t *store,
                          int width, int height,
                          int xstart, int ystart, int xsize, int ysize,
                          int x, int y, int key, int to_display)
{
    int sx, sy, dx, dy;
    uint8_t *storeline = store + ystart * width,
            *displayline = cfg->screen_start + y * cfg->line_length,
            pixel;

    /* Define the limits of the copying in the store. */

    int xlimit = xstart + xsize, ylimit = ystart + ysize;

    if (xlimit > width)
        xlimit = width;

    if (ylimit > height)
        ylimit = height;

    /* Perform the copying between the store and display. */

    for (sy = ystart, dy = y; (sy < ylimit) && (dy < cfg->line_count); sy++, dy++)
    {
        if (displayline >= cfg->screen_limit)
            displayline -= cfg->screen_size;

        for (sx = xstart, dx = x; (sx < xlimit) && (dx < cfg->line_length); sx++, dx++)
        {
            if (to_display)
            {
                pixel = storeline[sx];
                if ((key < 0) || (pixel != key))
                    displayline[get_position(cfg, dx)] = pixel;
            }
            else
                storeline[sx] = displayline[get_position(cfg, dx)];
        }

        storeline += width;
        displayline += cfg->line_length;
    }
}

/* Scroll the display. */

void scroll_display(display_config_t *cfg, int x, int y)
{
    uint8_t *start;

    /* Move the screen start by the given number of bytes and lines. */

    start = cfg->screen_start + x + y * cfg->line_length;

    /* Wrap around the start of the framebuffer to the end. */

    if (start < cfg->framebuffer)
        start = cfg->screen_limit - (cfg->screen_limit - start) % cfg->screen_size;

    /* Wrap around the end of the framebuffer to the start. */

    else if (start >= cfg->screen_limit)
        start = cfg->framebuffer + (start - cfg->framebuffer) % cfg->screen_size;

    cfg->screen_start = start;
}
